/*
	MarkovGen - A Markov Chain Program
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdexcept>
#include "arguments.hpp"
#include "util.hpp"

namespace
{
	enum
	{
		ALLOW_Az = 256
	};
}

Arguments::Arguments(int argc, char **argv) :
	min_match(1),
	max_match(1),
	min_generate(1),
	max_generate(1)
{
	argp_parse (&s_argp, argc, argv, 0, 0, this);
}

error_t Arguments::Parse(int key, char *arg, struct argp_state *state)
{
	switch (key)
	{
		case 'a':
			for (unsigned int i = 0; arg[i]; i++){
				allow.insert(arg[i]);
			}
			break;
		case 'f':
			for (unsigned int i = 0; arg[i]; i++){
				filter.insert(arg[i]);
			}
			break;
		case 'm':
			min_match = StringToInt(arg, state, 1);
			break;
		case 'n':
			max_match = StringToInt(arg, state, 1);
			break;
		case 'p':
			post_char = arg;
			break;
		case 'q':
			post_group = arg;
			break;
		case 's':
			for (unsigned int i = 0; arg[i]; i++){
				split.insert(arg[i]);
			}
			break;
		case 'x':
			min_generate = StringToInt(arg, state, 1);
			break;
		case 'y':
			max_generate = StringToInt(arg, state, 1);
			break;
		case ALLOW_Az:
			for (unsigned int i = 'A'; i <= 'Z'; i++){
				allow.insert(i);
				allow.insert(i + 'a' - 'A');
			}
			break;
		default:
			return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

const struct argp Arguments::s_argp = {Arguments::s_options, Arguments::parser, "A program to generate strings from Markov chains", 0, 0, 0};
const struct argp_option Arguments::s_options[] =
{
	{"allow", 'a', "CHARS", 0, "Allow only CHARS from the input (default all)", 0},
	{"filter", 'f', "CHARS", 0, "Remove CHARS from the input (default none)", 0},
	{"post-char", 'p', "STRING", 0, "Add STRING to the end of every generated character (default none)", 0},
	{"post-group", 'q', "STRING", 0, "Add STRING to the end of every generated character group (default none)", 0},
	{"split", 's', "CHARS", 0, "Split stream by any of CHARS (default none)", 0},
	{"min-match", 'm', "N", 0, "Try to match at least N characters (default 1)", 0},
	{"max-match", 'n', "N", 0, "Match at most N characters (default 1)", 0},
	{"min-generate", 'x', "N", 0, "Try to generate at least N characters (default 1)", 0},
	{"max-generate", 'y', "N", 0, "Generate at most N characters (default 1)", 0},
	
	{"allow-Az", ALLOW_Az, 0, 0, "Allow A-z", 0},
	{0}
};

error_t Arguments::parser(int key, char *arg, struct argp_state *state)
{
	Arguments *args = (Arguments *)state->input;
	return args->Parse(key, arg, state);
}

uint32_t Arguments::StringToInt(const char *arg, const struct argp_state *state, uint32_t min, uint32_t max)
{
	try
	{
		size_t pos = 0;
		long long n = std::stoll(arg, &pos, 0);
		if ((n >= min) && (n <= max) && !arg[pos]){
			return n;
		}
	}
	catch (std::invalid_argument e){}
	catch (std::out_of_range e){}
	
	argp_error (state, "Argument to %s is not an integer in range: %s", state->argv[state->arg_num], arg);
	throw FormatError("Return from argp_error()");
}
