/*
	MarkovGen - A Markov Chain Program
	Copyright (C) 2012-2013  Nicholas Tomlinson
	
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cinttypes>
#include "markovchain.hpp"
#include "util.hpp"

MarkovChain::MarkovChain(uint32_t max_match, uint32_t max_gen) :
	m_max_match(max_match), m_max_gen(max_gen)
{
}

void MarkovChain::ResetHistory()
{
	m_history.clear();
}

void MarkovChain::AddToHistory(const std::string &c)
{
	//Add to the front of the list (most recent first)
	m_history.push_front(c);
	
	//Don't allow the history to grow unnecessarily long
	if (m_history.size() > m_max_match + m_max_gen){
		m_history.pop_back();
	}
}

void MarkovChain::Record(const std::string &c)
{
	AddToHistory(c);
	
	//Update the chain data
	std::vector<std::string> gen;
	gen.reserve(m_max_gen);
	
	for (std::list<std::string>::iterator git = m_history.begin(); (git != m_history.end()) && (gen.size() < m_max_gen); git++){
		gen.push_back(*git); //Ensure the generator reflects the current
		
		//Iterate over the matches
		std::vector<std::string> match;
		match.reserve(m_max_match);
		RecordMatchGen(match, gen); //Record against the empty match
		
		//Copy the iterator so we can continue on where this left off, and use this to iterate
		std::list<std::string>::iterator mit = git;
		mit++;
		for (; (mit != m_history.end()) && (match.size() < m_max_match); mit++){
			match.push_back(*mit);
			RecordMatchGen(match, gen);
		}
	}
}

void MarkovChain::RecordMatchGen(const std::vector<std::string> &match, const std::vector<std::string> &gen)
{
	MapTarget &target = m_weights[match][gen.size()]; //Get the map corresponding to match
	if (!target.weights.count(gen)){
		target.weights[gen] = 0;
	}
	target.weights[gen]++;
	target.weight_sum++;
}

std::vector<std::string> MarkovChain::Generate(uint32_t min_match, uint32_t max_match, uint32_t min_gen, uint32_t max_gen)
{
	return GenerateMatchGen(min_match + (m_random()%(1 + max_match - min_match)), min_gen + (m_random()%(1 + m_max_gen - min_gen)));
}

std::vector<std::string> MarkovChain::GenerateMatchGen(uint32_t match_length, uint32_t gen_length)
{
	//Get a history of the correct length
	std::vector<std::string> history; //Copy the history so we can fiddle with it, as a vector
	history.reserve(match_length);
	for (const std::string &c: m_history){
		if (history.size() >= match_length){
			break;
		}
		history.push_back(c);
	}
	
	//Try to find a match
	while (history.size()){
		//We found a match with the appropriate generation length
		//Now we can generate a random number and return the corresponding string
		
		if (m_weights.count(history)){
			std::map<uint32_t, MapTarget> &weights_map = m_weights[history];
			if (weights_map.count(gen_length)){
				MapTarget &weights = weights_map[gen_length];
				
				uint64_t weight_sum = 0;
				uint64_t target = random() % (1 + weights.weight_sum);
				for (auto gen: weights.weights){
					weight_sum += gen.second;
					if (weight_sum >= target){
						const std::vector<std::string> &r = gen.first;
						for (std::vector<std::string>::const_reverse_iterator it = r.rbegin(); it != r.rend(); it++){
							AddToHistory(*it);
						}
						return r;
					}
				}
			}
		}
		history.pop_back(); //We didn't find a match with this length history, so try a shorter one
	}
	
	throw FormatError("Insufficient Markov data for a generation of length %" PRIu32, gen_length);
}
